import subprocess
from pathlib import Path
from typing import List, Literal, Optional

import attr

from mentat.errors import MentatError
from mentat.git_handler import (
    check_head_exists,
    get_diff_for_file,
    get_files_in_diff,
    get_git_root_for_path,
    get_treeish_metadata,
    get_untracked_files,
)
from mentat.interval import Interval
from mentat.session_context import SESSION_CONTEXT
from mentat.session_stream import SessionStream


@attr.define(frozen=True)
class DiffAnnotation(Interval):
    start: int | float = attr.field()
    message: List[str] = attr.field()
    end: int | float = attr.field(
        default=attr.Factory(
            lambda self: self.start + sum(bool(line.startswith("-")) for line in self.message),
            takes_self=True,
        )
    )


def parse_diff(diff: str) -> list[DiffAnnotation]:
    """Parse diff into a list of annotations."""
    annotations: list[DiffAnnotation] = []
    active_annotation: Optional[DiffAnnotation] = None
    lines = diff.splitlines()
    for line in lines:
        if line.startswith(("---", "+++", "//", "diff", "index")):
            continue
        elif line.startswith("@@"):
            if active_annotation:
                annotations.append(active_annotation)
            _new_index = line.split(" ")[2]
            if "," in _new_index:
                new_start = _new_index[1:].split(",")[0]
            else:
                new_start = _new_index[1:]
            active_annotation = DiffAnnotation(start=int(new_start), message=[])
        elif line.startswith(("+", "-")):
            if not active_annotation:
                raise MentatError("Invalid diff")
            active_annotation.message.append(line)
    if active_annotation:
        annotations.append(active_annotation)
    annotations.sort(key=lambda a: a.start)
    return annotations


def annotate_file_message(code_message: list[str], annotations: list[DiffAnnotation]) -> list[str]:
    """Return the code_message with annotations inserted."""
    active_index = 0
    annotated_message: list[str] = []
    for annotation in annotations:
        # Fill-in lines between annotations
        if active_index < annotation.start:
            unaffected_lines = code_message[active_index : annotation.start]
            annotated_message += unaffected_lines
        active_index = annotation.start
        if annotation.start == 0:
            # Make sure the PATH stays on line 1
            annotated_message.append(code_message[0])
            active_index += 1
        i_minus = None
        for line in annotation.message:
            sign = line[0]
            if sign == "+":
                # Add '+' lines in place of code_message lines
                annotated_message.append(f"{active_index}:{line}")
                active_index += 1
                i_minus = None
            elif sign == "-":
                # Insert '-' lines at the point they were removed
                i_minus = 0 if i_minus is None else i_minus
                annotated_message.append(f"{annotation.start + i_minus}:{line}")
                i_minus += 1
    if active_index < len(code_message):
        annotated_message += code_message[active_index:]

    return annotated_message


class DiffContext:
    def __init__(
        self,
        stream: SessionStream,
        cwd: Path,
        diff: Optional[str] = None,
        pr_diff: Optional[str] = None,
    ):
        self.git_root = get_git_root_for_path(cwd, raise_error=False)
        if not self.git_root:
            return

        if diff and pr_diff:
            # TODO: Once broadcast queue's unread messages and/or config is moved to client,
            # determine if this should quit or not
            stream.send(
                "Cannot specify more than one type of diff. Disabling diff and" " pr-diff.",
                style="warning",
            )
            diff = None
            pr_diff = None

        target = diff or pr_diff
        if not target:
            self.target = "HEAD"
            self.name = "HEAD (last commit)"
            return

        name = ""
        treeish_type = _get_treeish_type(self.git_root, target)
        if treeish_type is None:
            stream.send(f"Invalid treeish: {target}", style="failure")
            stream.send("Disabling diff and pr-diff.", style="warning")
            self.target = "HEAD"
            self.name = "HEAD (last commit)"
            return

        if treeish_type == "branch":
            name += f"Branch {target}: "
        elif treeish_type == "relative":
            name += f"{target}: "

        if pr_diff:
            name = f"Merge-base {name}"
            target = _git_command(self.git_root, "merge-base", "HEAD", pr_diff)
            if not target:
                # TODO: Same as above todo
                stream.send(
                    f"Cannot identify merge base between HEAD and {pr_diff}. Disabling" " pr-diff.",
                    style="warning",
                )
                self.target = "HEAD"
                self.name = "HEAD (last commit)"
                return

        meta = get_treeish_metadata(self.git_root, target)
        name += f'{meta["hexsha"][:8]}: {meta["summary"]}'
        if target == "HEAD":
            name = "HEAD (last commit)"

        self.target = target
        self.name = name

    _diff_files: List[Path] | None = None
    _untracked_files: List[Path] | None = None

    def diff_files(self) -> List[Path]:
        if not self.git_root:
            return []
        if self._diff_files is None:
            self.refresh()
        return self._diff_files  # pyright: ignore

    def untracked_files(self) -> List[Path]:
        if not self.git_root:
            return []
        if self._untracked_files is None:
            self.refresh()
        return self._untracked_files  # pyright: ignore

    def refresh(self):
        if not self.git_root:
            return
        ctx = SESSION_CONTEXT.get()

        if self.target == "HEAD" and not check_head_exists():
            self._diff_files = []  # A new repo without any commits
            self._untracked_files = []
        else:
            self._diff_files = [(ctx.cwd / f).resolve() for f in get_files_in_diff(self.target)]
            self._untracked_files = [(ctx.cwd / f).resolve() for f in get_untracked_files(ctx.cwd)]

    def get_annotations(self, rel_path: Path) -> list[DiffAnnotation]:
        if not self.git_root:
            return []
        diff = get_diff_for_file(self.target, rel_path)
        return parse_diff(diff)

    def get_display_context(self) -> Optional[str]:
        if not self.git_root:
            return None
        diff_files = self.diff_files()
        if not diff_files:
            return ""
        num_files = len(diff_files)
        num_lines = 0
        for file in diff_files:
            diff = get_diff_for_file(self.target, file)
            diff_lines = diff.splitlines()
            num_lines += len([line for line in diff_lines if line.startswith(("+ ", "- "))])
        return f" {self.name} | {num_files} files | {num_lines} lines"

    def annotate_file_message(self, rel_path: Path, file_message: list[str]) -> list[str]:
        """Return file_message annotated with active diff."""
        if not self.git_root:
            return []
        annotations = self.get_annotations(rel_path)
        return annotate_file_message(file_message, annotations)


TreeishType = Literal["commit", "branch", "relative"]


def _git_command(git_root: Path, *args: str) -> str | None:
    try:
        return subprocess.check_output(["git"] + list(args), cwd=git_root, stderr=subprocess.PIPE, text=True).strip()
    except subprocess.CalledProcessError:
        return None


def _get_treeish_type(git_root: Path, treeish: str) -> TreeishType | None:
    object_type = _git_command(git_root, "cat-file", "-t", treeish)

    if not object_type:
        return None

    if object_type == "commit":
        if "~" in treeish or "^" in treeish:
            return "relative"

        if _git_command(git_root, "show-ref", "--heads", treeish):
            return "branch"
        else:
            return "commit"
    return None



#!/usr/bin/env python
from __future__ import annotations

import asyncio
import importlib.util
import json
import os
import re
from datetime import datetime
from pathlib import Path
from uuid import uuid4

from openai.types.chat.completion_create_params import ResponseFormat

from benchmarks.arg_parser import common_benchmark_parser
from benchmarks.benchmark_result import BenchmarkResult
from benchmarks.benchmark_run import BenchmarkRun
from benchmarks.run_sample import run_sample
from benchmarks.swe_bench_runner import SWE_BENCH_SAMPLES_DIR, get_swe_samples
from mentat.config import Config
from mentat.git_handler import get_git_diff, get_mentat_branch, get_mentat_hexsha
from mentat.llm_api_handler import model_context_size, prompt_tokens
from mentat.sampler.sample import Sample
from mentat.sampler.utils import setup_repo
from mentat.session_context import SESSION_CONTEXT


def git_diff_from_comparison_commit(sample: Sample, comparison_commit: str) -> str:
    starting_cwd = Path.cwd()
    repo = setup_repo(
        url=sample.repo,
        cwd=None,
        commit=sample.merge_base,
        diff_merge_base=sample.diff_merge_base,
        diff_active=sample.diff_active,
    )
    cwd = Path(repo.working_dir)
    diff = get_git_diff("HEAD", comparison_commit, cwd=cwd)
    os.chdir(starting_cwd)
    return diff


async def grade(to_grade, prompt, model="gpt-4-1106-preview"):
    try:
        messages = [
            {"role": "system", "content": prompt},
            {"role": "user", "content": to_grade},
        ]
        tokens = prompt_tokens(messages, model)
        max_tokens = model_context_size(model) - 1000  # Response buffer
        if tokens > max_tokens:
            print("Prompt too long! Truncating... (this may affect results)")
            tokens_to_remove = tokens - max_tokens
            chars_per_token = len(str(messages)) / tokens
            chars_to_remove = int(chars_per_token * tokens_to_remove)
            messages[1]["content"] = messages[1]["content"][:-chars_to_remove]

        llm_api_handler = SESSION_CONTEXT.get().llm_api_handler
        llm_grade = await llm_api_handler.call_llm_api(messages, model, False, ResponseFormat(type="json_object"))
        content = llm_grade.choices[0].message.content
        return json.loads(content)
    except Exception as e:
        return {"error": str(e)}


syntax_grading_prompt = """\
You will be given a git diff that was generated by an automated system. Your job
is to flag certain common errors. Please reply with a json object with the
following schema:
off_by_one: true if you believe a line was inserted at the wrong place otherwise
false
The following two fields are only required if off_by_one is true:
off_by_one_lines: a list of line numbers that you believe were inserted at the
wrong place
off_by_one_direction: a list of integers that are how off you believe the
insertions were. A positive number means the line was inserted too low, a
negative numbers means to high.
indentation: true if you believe the indentation is incorrect otherwise false
The following two fields are only required if indentation is true:
indentation_lines: a list of line numbers that you believe have incorrect
indentation.
indentation_direction: a list of integers that are how off you believe the
indentation is. A positive number means the line was indented too far, a
negative number means not enough.
syntax: true if you believe there is a syntax error unrelated to insertion
location or indentation.
syntax_description: a string describing the syntax errors if present."""


async def grade_diff_syntax(diff):
    return await grade(diff, syntax_grading_prompt)


model_response_grade_prompt = """\
You will be give a models response to a prompt. You won't be given the full
context of the response. You are just looking for certain stylistic errors.
Respond in json. The following fields are required:
referenced_format: boolean, true if the model talks about its edit format in any
way in its response. For example if it has a clause like "The edits in the
requested format are:"
trailing_waffling: boolean, true if after the structured edits the model ends
with a clause like "Please note I may not have had all the information I needed" """


async def grade_model_response(model_response):
    return await grade(model_response, model_response_grade_prompt)


comparison_prompt = """\
You will be given two diffs. The first was human written and the second was
generated by an automated system. Your job is to grade the automated diff. Repond in
json. The following fields are required:
missing_functionality: true if the generated diff is missing functionality
present in the human written pr.
missing_description: optional string describing what's missing
extra_functionality: true if the generated diff has functionality not present
in the human written pr.
extra_description: optional string describing what's extra"""


async def compare_diffs(actual, generated):
    prompt = f"HUMAN WRITTEN DIFF:\n{actual}\nGENERATED DIFF:\n{generated}"

    return await grade(prompt, comparison_prompt)


async def grade_diff(diff, response, result, comparison_diff=None):
    # Set syntax and response grade information
    result.code = diff
    diff_grade = await grade_diff_syntax(diff)
    result.diff_grade = diff_grade
    result.off_by_one = diff_grade.get("off_by_one")
    result.indentation_error = diff_grade.get("indentation")
    result.syntax_error = diff_grade.get("syntax")
    response_grade = await grade_model_response(response)
    result.response_grade = response_grade
    result.referenced_format = response_grade.get("referenced_format")

    # Set comparison grade information
    if comparison_diff:
        comparison_grade = await compare_diffs(diff, comparison_diff)
        result.comparison_grade = comparison_grade
        result.extra_functionality = comparison_grade.get("extra_functionality")
        result.missing_functionality = comparison_grade.get("missing_functionality")

    return result


class Benchmark:
    def __init__(
        self,
        title: str,
        description: str = "",
        config: Config = Config(),
        verify: callable | None = None,
        samples: list[Sample] = [],
    ):
        self.title = title
        self.description = description
        self.config = config
        self.verify = verify
        self.samples = samples

    @classmethod
    def from_module(cls, path_to_module: Path, module_name: str) -> Benchmark:
        # Dynamic import
        spec = importlib.util.spec_from_file_location(module_name, path_to_module)
        assert spec and spec.loader
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        output = cls(
            title=module.title,
            description=module.description,
            config=module.config,
            verify=module.verify if hasattr(module, "verify") else None,
            samples=[
                # Create new samples for each prompt
                Sample(
                    title=module.title,
                    description=module.description,
                    id="",
                    parent_id="",
                    repo=module.repo,
                    merge_base=module.commit,
                    diff_merge_base="",
                    diff_active="",
                    message_history=[],
                    message_prompt=prompt,
                    message_edit="",
                    context=getattr(module, "minimum_context", []),
                    diff_edit="",
                )
                for prompt in module.prompts
            ],
        )
        if hasattr(module, "comparison_commit"):
            diff_edit = git_diff_from_comparison_commit(output.samples[0], module.comparison_commit)
            for sample in output.samples:
                if not sample.diff_edit:
                    sample.diff_edit = diff_edit
        return output

    @classmethod
    def from_sample(cls, path_to_sample: Path) -> Benchmark:
        sample = Sample.load(path_to_sample)
        return cls(
            title=sample.title,
            description=sample.description,
            samples=[sample],
        )

    async def run(self, retries: int = 1) -> list[BenchmarkResult]:
        print("Benchmark:", self.title)
        start_dir = Path.cwd()
        results: list[BenchmarkResult] = []
        for i, sample in enumerate(self.samples):
            print("  Prompt:", sample.message_prompt)
            for j in range(1, retries + 1):
                formatted_title = re.sub(r"[ '\"/\\-^]", "", sample.title).replace(" ", "_")
                result = BenchmarkResult(
                    name=f"{formatted_title}-{i}-{j}",
                    family=formatted_title,
                )
                try:
                    sample_result = await run_sample(sample)
                    result.cost = sample_result["cost"]
                    result.tokens = sample_result["tokens"]
                    result.transcript = sample_result["transcript"]
                    if self.verify is not None:
                        result.verify = self.verify()

                    await grade_diff(
                        sample_result["diff_eval"],
                        sample_result["message_eval"],
                        result,
                        sample.diff_edit,
                    )
                except Exception as e:
                    result.run_error = str(e)
                finally:
                    results.append(result)
                    os.chdir(start_dir)
        return results


def benchmark_listed(title, benchmarks):
    for b in benchmarks:
        if b.lower() in title.lower():
            return True
    return False


def run_benchmarks(user_benchmarks: list[str], directory: str, retries: int = 1):
    # Load benchmarks
    dir_path = Path(directory).resolve()
    assert dir_path.exists(), f"Invalid directory: {directory}"
    print(f"Running benchmarks from {dir_path}")
    benchmarks: list[Benchmark] = []
    for root, dirs, files in os.walk(dir_path):
        for file in files:
            path = Path(root) / file
            if file.endswith(".py"):
                benchmark = Benchmark.from_module(path, "benchmark")
            elif file.endswith(".json"):
                benchmark = Benchmark.from_sample(path)
            else:
                continue

            if len(user_benchmarks) > 0 and not benchmark_listed(benchmark.title, user_benchmarks):
                continue
            benchmarks.append(benchmark)
    print("Found benchmarks:\n" + "\n".join(b.title for b in benchmarks))
    print("*" * 80)

    # Run benchmarks
    results_cache = dir_path / f"benchmark_results_cache_{uuid4()}.jsonl"
    results_cache.touch()
    total_cost = 0.0
    for benchmark in benchmarks:
        # Run benchmark.run() with timeout
        try:
            result = asyncio.run(benchmark.run(retries=retries))
            with open(results_cache, "a") as f:
                for r in result:
                    total_cost += r.cost if r.cost else 0.0
                    f.write(r.to_json() + "\n")
        except KeyboardInterrupt:
            # TODO: Prints none on first ctrl+c, then here - probably the PythonClient
            print("Exiting...")
            break
        except Exception as e:
            print(f"Error running benchmark {benchmark.title}: {e}")
            continue

    # Summarize results
    print(f"Total cost: {total_cost}")
    with open(results_cache, "r") as f:
        results = [BenchmarkResult.load_json(line) for line in f.readlines()]
    benchmark_run = BenchmarkRun(
        results,
        metadata={
            "type": "Sampled",
            "date": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "commit": get_mentat_hexsha(),
            "branch": get_mentat_branch(),
        },
    )
    benchmark_run.save()

    results_cache.unlink()  # Delete cache
    benchmark_run.render_results()


if __name__ == "__main__":
    parser = common_benchmark_parser()
    args = parser.parse_args()
    if args.swe_bench:
        if args.swe_bench not in {"dev", "train", "test"}:
            print("Invalid SWE-Bench split.")
            exit(1)
        # Download and save SWE benchmarks as Samples
        samples = get_swe_samples(args.swe_bench, args.max_benchmarks)
        sample_titles = [sample.title for sample in samples]
        args.benchmarks = sample_titles
        args.directory = SWE_BENCH_SAMPLES_DIR / args.swe_bench
    run_benchmarks(
        args.benchmarks,
        args.directory,
        args.retries,
    )



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING, List

from mentat.edit_history import EditHistory
from mentat.errors import MentatError
from mentat.interval import Interval
from mentat.session_context import SESSION_CONTEXT
from mentat.session_input import ask_yes_no
from mentat.utils import get_relative_path, sha256

if TYPE_CHECKING:
    # This normally will cause a circular import
    from mentat.parsers.file_edit import FileEdit


class CodeFileManager:
    def __init__(self):
        self.file_lines = dict[Path, list[str]]()
        self.history = EditHistory()

    def read_file(self, path: Path) -> list[str]:
        # TODO: Change to only ever using this function to read files, then cache files and
        # only re-read them when their last modified time is updated
        session_context = SESSION_CONTEXT.get()

        abs_path = path if path.is_absolute() else session_context.cwd / path
        with open(abs_path, "r") as f:
            lines = f.read().split("\n")
        self.file_lines[abs_path] = lines
        return lines

    def create_file(self, abs_path: Path, content: str = ""):
        ctx = SESSION_CONTEXT.get()
        code_context = ctx.code_context

        logging.info(f"Creating new file {abs_path}")
        # Create any missing directories in the path
        abs_path.parent.mkdir(parents=True, exist_ok=True)
        with open(abs_path, "w") as f:
            f.write(content)

        if abs_path not in code_context.include_files:
            code_context.include(abs_path)

    def delete_file(self, abs_path: Path):
        ctx = SESSION_CONTEXT.get()
        code_context = ctx.code_context

        logging.info(f"Deleting file {abs_path}")

        if abs_path in code_context.include_files:
            code_context.exclude(abs_path)
        abs_path.unlink()

    def rename_file(self, abs_path: Path, new_abs_path: Path):
        ctx = SESSION_CONTEXT.get()
        code_context = ctx.code_context

        logging.info(f"Renaming file {abs_path} to {new_abs_path}")
        if abs_path in code_context.include_files:
            code_context.exclude(abs_path)
        os.rename(abs_path, new_abs_path)
        if new_abs_path not in code_context.include_files:
            code_context.include(new_abs_path)

    def write_to_file(self, abs_path: Path, new_lines: List[str]):
        with open(abs_path, "w") as f:
            f.write("\n".join(new_lines))
        self.file_lines[abs_path] = new_lines

    # Mainly does checks on if file is in context, file exists, file is unchanged, etc.
    async def write_changes_to_files(
        self,
        file_edits: list[FileEdit],
    ) -> list[FileEdit]:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        agent_handler = session_context.agent_handler

        if not file_edits:
            return []

        applied_edits: list[FileEdit] = []
        for file_edit in file_edits:
            display_path = get_relative_path(file_edit.file_path, session_context.cwd)

            if file_edit.is_creation:
                if file_edit.file_path.exists():
                    raise MentatError(f"Model attempted to create file {file_edit.file_path} which" " already exists")
                self.create_file(file_edit.file_path)
            elif not file_edit.file_path.exists():
                raise MentatError(f"Attempted to edit non-existent file {file_edit.file_path}")

            if file_edit.is_deletion:
                stream.send(f"Deleting {display_path}...", style="error")
                # We use the current lines rather than the stored lines for undo
                file_edit.previous_file_lines = self.read_file(file_edit.file_path)
                self.delete_file(file_edit.file_path)
                applied_edits.append(file_edit)
                continue

            if not file_edit.is_creation:
                # TODO: We use read_file so much that this probably doesn't work anymore
                # We should instead make sure the last modified time doesn't change between code message and now
                stored_lines = self.file_lines[file_edit.file_path]
                if stored_lines != self.read_file(file_edit.file_path):
                    logging.info(f"File '{file_edit.file_path}' changed while generating changes")
                    stream.send(
                        f"File '{display_path}' changed while"
                        " generating; current file changes will be erased. Continue?",
                        style="warning",
                    )
                    if not await ask_yes_no(default_yes=False):
                        stream.send(f"Not applying changes to file {display_path}")
                        continue
            else:
                stored_lines = []

            if file_edit.rename_file_path is not None:
                if file_edit.rename_file_path.exists():
                    raise MentatError(
                        f"Attempted to rename file {file_edit.file_path} to existing"
                        f" file {file_edit.rename_file_path}"
                    )
                self.rename_file(file_edit.file_path, file_edit.rename_file_path)

            new_lines = file_edit.get_updated_file_lines(stored_lines)
            if new_lines != stored_lines:
                file_path = file_edit.rename_file_path or file_edit.file_path
                # We use the current lines rather than the stored lines for undo
                file_edit.previous_file_lines = self.read_file(file_path)
                self.write_to_file(file_path, new_lines)
            applied_edits.append(file_edit)

        for applied_edit in applied_edits:
            self.history.add_edit(applied_edit)
        if not agent_handler.agent_enabled:
            self.history.push_edits()
        return applied_edits

    def get_file_checksum(self, path: Path, interval: Interval | None = None) -> str:
        if path.is_dir():
            return ""  # TODO: Build and maintain a hash tree for git_root
        text = path.read_text()
        if interval is not None:
            lines = text.splitlines()
            filtered_lines = [line for i, line in enumerate(lines, start=1) if interval.contains(i)]
            text = "\n".join(filtered_lines)
        return sha256(text)



#!/usr/bin/env python
import asyncio
import json
import os
from collections import defaultdict
from itertools import islice
from pathlib import Path
from typing import Any

from git import Repo

from benchmarks.arg_parser import common_benchmark_parser
from mentat.code_context import CodeContext
from mentat.code_feature import CodeFeature
from mentat.code_file_manager import CodeFileManager
from mentat.config import Config
from mentat.cost_tracker import CostTracker
from mentat.llm_api_handler import count_tokens, model_context_size
from mentat.sampler.utils import clone_repo
from mentat.session_context import SESSION_CONTEXT, SessionContext


class MockStream:
    def send(self, message, **kwargs):
        end = kwargs.get("end", "\n")
        print(message, end=end)


def _load_benchmarks() -> dict[str, dict[str, Any]]:
    """Load all benchmarks found in benchmark_repos"""
    benchmarks = {}
    benchmarks_dir = Path(__file__).parent / "../benchmark_repos"
    for repo_dir in benchmarks_dir.iterdir():
        benchmarks_path = repo_dir / "benchmarks.json"
        if benchmarks_path.exists():
            with open(benchmarks_path, "r") as f:
                benchmarks.update(json.load(f))
    return benchmarks


def _convert_features_to_line_sets(git_root: Path, features: list[CodeFeature]) -> defaultdict[set]:
    """Convert a list of features to a dict of {path: set(lines)} for comparison"""
    lines = defaultdict(set)
    for feature in features:
        # Non-explicit features (e.g. CodeMaps) are considered false positives.
        # Using negative numbers here as that affect.

        path = feature.path.relative_to(git_root)
        interval = feature.interval
        lines[path].update(range(interval.start, interval.end + 1))
    return lines


def evaluate(
    git_root: Path,
    actual: list[CodeFeature],
    expected: list[CodeFeature],
) -> dict[str, float]:
    """Compare two lists of features and return precision, recall and f1 scores"""
    actual_lines = _convert_features_to_line_sets(git_root, actual)
    expected_lines = _convert_features_to_line_sets(git_root, expected)

    _TP, _FP, _FN = 0, 0, 0
    for file in actual_lines | expected_lines:
        actual_set = actual_lines[file]
        expected_set = expected_lines[file]
        _TP += len(actual_set & expected_set)
        _FP += len(actual_set - expected_set)
        _FN += len(expected_set - actual_set)

    precision, recall, f1 = None, None, None
    if (_TP + _FP) > 0:
        precision = _TP / (_TP + _FP)
    if (_TP + _FN) > 0:
        recall = _TP / (_TP + _FN)
    if precision and recall:
        f1 = 2 * precision * recall / (precision + recall)

    return {"precision": precision, "recall": recall, "f1": f1}


async def select_features_for_benchmark(session_context, benchmark, eval=True, use_expected=False, use_llm=True):
    """Select features for benchmark using expected edits as a guide"""
    git_root = session_context.git_root
    config = session_context.config
    parser = config.parser
    code_context = session_context.code_context

    # The longest context that could have been included to generate expected_edits
    model = config.model
    mentat_prompt_tokens = count_tokens(parser.get_system_prompt(), model)
    expected_edits, expected_edits_tokens = None, 0
    if use_expected:
        expected_edits = benchmark["expected_edits"]
        expected_edits_tokens = count_tokens(expected_edits, model)
    max_context_tokens = model_context_size(model) - mentat_prompt_tokens - expected_edits_tokens
    # Fill-in available context
    config.auto_context_tokens = 8000
    code_context.use_llm = use_llm
    await code_context.get_code_message(benchmark["prompt"], max_context_tokens, expected_edits)
    git_root_length = len(str(git_root)) + 1
    selected_features = [f.ref()[git_root_length:] for f in code_context.features]

    selector_performance = {}
    if eval:
        edited_features = [CodeFeature(git_root / f) for f in benchmark["edited_features"]]
        selector_performance = evaluate(git_root, code_context.features, edited_features)
    return {"features": selected_features, "score": selector_performance}


async def test_code_context_performance(benchmarks, max_benchmarks=10):
    """Run a set of benchmarks and evaluate performance

    Run standalone:
        `./benchmarks/context_benchmark.py`
    """
    # Load applicable benchmarks
    all_benchmarks = _load_benchmarks()
    if len(benchmarks) > 0:
        benchmarks_to_run = {k: v for k, v in all_benchmarks.items() if k in benchmarks}
    else:
        benchmarks_to_run = dict(islice(all_benchmarks.items(), max_benchmarks))

    # Run each one
    scores = {}
    for benchmark in benchmarks_to_run.values():
        print("\n" + benchmark["prompt"])

        # Setup the cwd the same way as in generate
        url = benchmark["codebase_url"]
        codebase = clone_repo(url=url, local_dir_name=url.split("/")[-1], refresh=False)
        os.chdir(codebase)
        repo = Repo(".")
        repo.git.checkout(benchmark["commit"])

        # Initialize a full SESSION_CONTEXT
        stream = MockStream()
        config = Config()
        code_context = CodeContext(stream, os.getcwd())
        session_context = SessionContext(
            stream,
            CostTracker(),
            Path.cwd(),
            config,
            code_context,
            CodeFileManager(),
            None,
        )
        SESSION_CONTEXT.set(session_context)

        # Run the benchmark and print results
        scores = []
        for use_llm in [False, True]:
            for use_expected in [False, True]:
                try:
                    if not use_llm and use_expected:
                        continue  # Not relevant
                    results = await select_features_for_benchmark(
                        session_context,
                        benchmark,
                        eval=True,
                        use_expected=use_expected,
                        use_llm=use_llm,
                    )
                    score = {
                        **results["score"],
                        "selected_features": results["features"],
                        "edited_features": benchmark["edited_features"],
                        "use_llm": use_llm,
                        "use_expected": use_expected,
                    }
                    scores.append(score)
                    print(
                        f"  UseExpected={use_expected}\t"
                        f"| LLM={use_llm}\t"
                        f"| Recall={(score['recall'] or 0.):.3f}\t"
                        f"| Precision={(score['precision'] or 0.):.3f}"
                    )
                except Exception as e:
                    print(f"Error: '{e}'; skipping")

    return scores


if __name__ == "__main__":
    parser = common_benchmark_parser()
    args = parser.parse_args()
    asyncio.run(
        test_code_context_performance(
            args.benchmarks,
            args.max_benchmarks,
        )
    )



import os
from pathlib import Path

import pkg_resources
from setuptools import find_packages, setup

readme_path = os.path.join(Path(__file__).parent, "README.md")
with open(readme_path, "r", encoding="utf-8") as f:
    long_description = f.read()


version_path = os.path.join(Path(__file__).parent, "mentat/VERSION")
with open(version_path, "r", encoding="utf-8") as f:
    VERSION = f.read().strip()

setup(
    name="mentat",
    version=VERSION,
    python_requires=">=3.10",
    packages=find_packages(include=["mentat", "mentat.*", "benchmarks", "benchmarks.*"]),
    install_requires=[
        str(r)
        for r in pkg_resources.parse_requirements(open(os.path.join(os.path.dirname(__file__), "requirements.txt")))
    ],
    entry_points={
        "console_scripts": [
            "mentat=mentat.terminal.client:run_cli",
            "mentat-server=mentat.server.mentat_server:main",
            "mentat-daemon=mentat.daemon:main",
        ],
    },
    description="AI coding assistant on your command line",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    include_package_data=True,
    extras_require={
        "dev": [
            str(r)
            for r in pkg_resources.parse_requirements(
                open(os.path.join(os.path.dirname(__file__), "dev-requirements.txt"))
            )
        ],
    },
)



import asyncio
import queue
from asyncio import Event
from timeit import default_timer
from typing import Any, List

import numpy as np
from typing_extensions import override

try:
    import sounddevice as sd
    import soundfile as sf

    audio_available = True
except Exception:
    audio_available = False


from mentat.command.command import Command, CommandArgument
from mentat.logging_config import logs_path
from mentat.session_context import SESSION_CONTEXT

RATE = 16000


class Recorder:
    def __init__(self):
        self.shutdown = Event()
        (logs_path / "audio").mkdir(parents=True, exist_ok=True)

        self.file = logs_path / "audio/talk_transcription.wav"

    def callback(
        self,
        in_data: np.ndarray[Any, Any],
        frames: int,
        time: Any,
        status: int,
    ):
        self.q.put(in_data.copy())

    async def record(self):
        self.start_time = default_timer()

        self.q: queue.Queue[np.ndarray[Any, Any]] = queue.Queue()
        with sf.SoundFile(  # pyright: ignore[reportUnboundVariable]
            self.file, mode="w", samplerate=RATE, channels=1
        ) as file:
            with sd.InputStream(  # pyright: ignore[reportUnboundVariable]
                samplerate=RATE, channels=1, callback=self.callback
            ):
                while not self.shutdown.is_set():
                    await asyncio.sleep(0)
                    file.write(self.q.get())  # type: ignore

        self.recording_time = default_timer() - self.start_time


class TalkCommand(Command, command_name="talk"):
    @override
    async def apply(self, *args: str) -> None:
        ctx = SESSION_CONTEXT.get()
        if not audio_available:
            # sounddevice manages port audio on Mac and Windows so we print an apt specific message
            ctx.stream.send(
                "Audio is not available on this system. You probably need to install"
                " PortAudio. For example `sudo apt install libportaudio2` on Ubuntu.",
                style="error",
            )
        else:
            # TODO: Ctrl+C doesn't make sense for VSCode client. Send this info in a client agnostic way
            ctx.stream.send("Listening on your default microphone. Press Ctrl+C to end.")
            recorder = Recorder()
            async with ctx.stream.interrupt_catcher(recorder.shutdown):
                await recorder.record()
            ctx.stream.send("Processing audio with whisper...")
            await asyncio.sleep(0.01)
            transcript = await ctx.llm_api_handler.call_whisper_api(recorder.file)
            ctx.stream.send(transcript, channel="default_prompt")
            ctx.cost_tracker.log_whisper_call_stats(recorder.recording_time)

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return []

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return []

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Start voice to text."



import argparse
import asyncio
from io import TextIOWrapper
import logging
from asyncio import CancelledError, Event
from pathlib import Path

from mentat.config import Config
from mentat.session import Session
from mentat.session_stream import StreamMessage, StreamMessageSource


async def ainput(fd_input: TextIOWrapper):
    return await asyncio.to_thread(fd_input.readline)


class MentatServer:
    def __init__(self, cwd: Path, config: Config) -> None:
        self.cwd = cwd
        self.stopped = Event()
        self.session = Session(self.cwd, config=config, apply_edits=False)

    async def _client_listener(self):
        with open(3) as fd_input:
            try:
                while not self.stopped.is_set():
                    line = await ainput(fd_input)
                    try:
                        message = StreamMessage.model_validate_json(line)  # pyright: ignore
                        self.session.stream.send_stream_message(message)
                    except ValueError:
                        logging.error(f"Invalid StreamMessage recieved: {line}")
            except Exception as e:
                logging.error(f"Error: {e}")

    async def _stream_listener(self):
        with open(4, "w") as fd_output:
            async for message in self.session.stream.universal_listen():
                if message.source == StreamMessageSource.SERVER:
                    message_json = StreamMessage.model_dump_json(message)
                    fd_output.write(message_json + "\n")
                    fd_output.flush()
                elif message.channel == "session_exit":
                    self.stopped.set()
                    break

    async def run(self):
        self.session.start()
        logging.debug("Completed startup")

        stream_listener_task = asyncio.create_task(self._stream_listener())
        client_listener_task = asyncio.create_task(self._client_listener())
        await self.stopped.wait()

        try:
            stream_listener_task.cancel()
            client_listener_task.cancel()
        except CancelledError:
            pass


async def run(args: argparse.Namespace):
    try:
        cwd = Path(args.cwd).expanduser().resolve()
        config = Config.create(cwd, args)
        mentat_server = MentatServer(cwd, config)
        await mentat_server.run()
    except Exception as e:
        logging.error(f"Exception: {e}")


def main():
    parser = argparse.ArgumentParser(
        description="Run conversation with command line args",
    )
    parser.add_argument("cwd", help="The working directory for the server to run in")
    Config.add_fields_to_argparse(parser)

    args = parser.parse_args()
    asyncio.run(run(args))



from mentat.parsers.block_parser import BlockParser
from mentat.parsers.json_parser import JsonParser
from mentat.parsers.parser import Parser
from mentat.parsers.replacement_parser import ReplacementParser
from mentat.parsers.unified_diff_parser import UnifiedDiffParser

parser_map: dict[str, Parser] = {
    "block": BlockParser(),
    "replacement": ReplacementParser(),
    "unified-diff": UnifiedDiffParser(),
    # JsonParser is experimental and has no streaming or tests and seems worse than other formats
    # Don't use it! But if you need to, make sure to set to a model that can use the JSON response_format
    "json": JsonParser(),
}



# ruff: noqa: F401
# type: ignore

# Import all of the commands so that they are initialized
from .agent import AgentCommand
from .amend import AmendCommand
from .clear import ClearCommand
from .commit import CommitCommand
from .config import ConfigCommand
from .exclude import ExcludeCommand
from .help import HelpCommand
from .include import IncludeCommand
from .load import LoadCommand
from .redo import RedoCommand
from .run import RunCommand
from .sample import SampleCommand
from .save import SaveCommand
from .screenshot import ScreenshotCommand
from .search import SearchCommand
from .talk import TalkCommand
from .undo import UndoCommand
from .undoall import UndoAllCommand
from .viewer import ViewerCommand



from __future__ import annotations

import argparse
import asyncio
import logging
from asyncio import Event
from pathlib import Path
from typing import Any, Coroutine, List, Set

from textual.app import AutopilotCallbackType

from mentat.code_context import ContextStreamMessage
from mentat.config import Config
from mentat.session import Session
from mentat.session_stream import StreamMessageSource
from mentat.terminal.terminal_app import TerminalApp


class TerminalClient:
    def __init__(
        self,
        cwd: Path = Path.cwd(),
        paths: List[str] = [],
        exclude_paths: List[str] = [],
        ignore_paths: List[str] = [],
        diff: str | None = None,
        pr_diff: str | None = None,
        config: Config = Config(),
        # Used for testing
        headless: bool = False,
        auto_pilot: AutopilotCallbackType | None = None,
    ):
        self.cwd = cwd
        self.paths = [Path(path) for path in paths]
        self.exclude_paths = [Path(path) for path in exclude_paths]
        self.ignore_paths = [Path(path) for path in ignore_paths]
        self.diff = diff
        self.pr_diff = pr_diff
        self.config = config
        self.headless = headless
        self.auto_pilot = auto_pilot

        self._tasks: Set[asyncio.Task[None]] = set()
        self._should_exit = Event()
        self._stopped = Event()

    def _create_task(self, coro: Coroutine[None, None, Any]):
        """Utility method for running a Task in the background"""

        def task_cleanup(task: asyncio.Task[None]):
            if task.exception() is not None:
                self.session.stream.send(
                    f"Error in task {task.get_coro()}: {str(task.exception())}",
                    style="error",
                )
                self._should_exit.set()
            self._tasks.remove(task)

        task = asyncio.create_task(coro)
        task.add_done_callback(task_cleanup)
        self._tasks.add(task)

        return task

    async def _run_terminal_app(self):
        self.app = TerminalApp(self)
        await self.app.run_async(headless=self.headless, auto_pilot=self.auto_pilot)
        self._should_exit.set()
        asyncio.create_task(self._shutdown())

    async def _default_channel_stream(self):
        async for message in self.session.stream.listen():
            self.app.display_stream_message(message)

    async def _default_prompt_stream(self):
        self._default_prompt = ""
        async for message in self.session.stream.listen("default_prompt"):
            self._default_prompt += message.data

    async def _listen_for_context_updates(self):
        async for message in self.session.stream.listen("context_update"):
            data: ContextStreamMessage = message.data
            (
                cwd,
                diff_context_display,
                auto_context_tokens,
                features,
                git_diff_paths,
                git_untracked_paths,
                total_tokens,
                total_cost,
            ) = (
                Path(data["cwd"]),
                data["diff_context_display"],
                data["auto_context_tokens"],
                data["features"],
                set(Path(path) for path in data["git_diff_paths"]),
                set(Path(path) for path in data["git_untracked_paths"]),
                data["total_tokens"],
                data["total_cost"],
            )
            self.app.update_context(
                cwd,
                diff_context_display,
                auto_context_tokens,
                features,
                git_diff_paths,
                git_untracked_paths,
                total_tokens,
                total_cost,
            )

    async def _handle_loading_messages(self):
        async for message in self.session.stream.listen("loading"):
            if message.extra.get("terminate", False):
                self.app.end_loading()
            else:
                self.app.start_loading()

    async def _handle_input_requests(self):
        while True:
            input_request_message = await self.session.stream.recv("input_request")

            default_prompt = self._default_prompt.strip()
            self._default_prompt = ""

            command_autocomplete = input_request_message.extra.get("command_autocomplete", False)

            user_input = await self.app.get_user_input(
                default_prompt=default_prompt,
                command_autocomplete=command_autocomplete,
            )

            self.session.stream.send(
                user_input,
                source=StreamMessageSource.CLIENT,
                channel=f"input_request:{input_request_message.id}",
            )

    async def _listen_for_session_stopped(self):
        await self.session.stream.recv(channel="session_stopped")
        self.app.disable_app()

    async def _listen_for_client_exit(self):
        await self.session.stream.recv(channel="client_exit")
        asyncio.create_task(self._shutdown())

    async def _listen_for_should_exit(self):
        """
        This listens for a user event signaling session shutdown (like an error), and tells the session to shutdown.
        Does *NOT* shut down the client, only disables it.
        """
        await self._should_exit.wait()
        self.session.stream.send(None, source=StreamMessageSource.CLIENT, channel="session_exit")

    def send_interrupt(self):
        if (
            # If session is still starting up we want to quit without an error
            not self.session or self.session.stream.is_interrupt_locked() is False or self.session.stopped.is_set()
        ):
            if self._should_exit.is_set():
                logging.debug("Force exiting client...")
                exit(0)
            else:
                logging.debug("Should exit client...")
                asyncio.create_task(self._shutdown())
                self._should_exit.set()
        else:
            logging.debug("Sending interrupt to session stream")
            self.session.stream.send(None, source=StreamMessageSource.CLIENT, channel="interrupt")

    async def _run(self):
        self.session = Session(
            self.cwd,
            self.paths,
            self.exclude_paths,
            self.ignore_paths,
            self.diff,
            self.pr_diff,
            self.config,
        )
        self.session.start()

        self._create_task(self._run_terminal_app())
        self._create_task(self._default_channel_stream())
        self._create_task(self._handle_input_requests())
        self._create_task(self._listen_for_context_updates())
        self._create_task(self._handle_loading_messages())
        self._create_task(self._default_prompt_stream())
        self._create_task(self._listen_for_client_exit())
        self._create_task(self._listen_for_session_stopped())
        self._create_task(self._listen_for_should_exit())

        logging.debug("Completed startup")
        await self._stopped.wait()

    async def _shutdown(self):
        logging.debug("Running shutdown")
        await self.session.stopped.wait()

        # Stop all background tasks
        for task in self._tasks:
            task.cancel()
        self._stopped.set()

    def run(self):
        asyncio.run(self._run())


def get_parser():
    parser = argparse.ArgumentParser(description="Run conversation with command line args")
    parser.add_argument(
        "paths",
        nargs="*",
        default=[],
        help="List of file paths, directory paths, or glob patterns",
    )
    parser.add_argument(
        "--exclude",
        "-e",
        nargs="*",
        default=[],
        help="List of file paths, directory paths, or glob patterns to exclude",
    )
    parser.add_argument(
        "--ignore",
        "-g",
        nargs="*",
        default=[],
        help=("List of file paths, directory paths, or glob patterns to ignore in" " auto-context"),
    )
    parser.add_argument(
        "--diff",
        "-d",
        nargs="?",
        type=str,
        default=None,
        const="HEAD",
        help="A git tree-ish (e.g. commit, branch, tag) to diff against",
    )
    parser.add_argument(
        "--pr-diff",
        "-p",
        type=str,
        default=None,
        help="A git tree-ish to diff against the latest common ancestor of",
    )
    parser.add_argument("--cwd", default=Path.cwd(), help="The current working directory")
    Config.add_fields_to_argparse(parser)
    return parser


def run_cli():
    parser = get_parser()

    args = parser.parse_args()

    cwd = Path(args.cwd).expanduser().resolve()
    paths = args.paths
    exclude_paths = args.exclude
    ignore_paths = args.ignore
    diff = args.diff
    pr_diff = args.pr_diff

    config = Config.create(cwd, args)

    terminal_client = TerminalClient(
        cwd,
        paths,
        exclude_paths,
        ignore_paths,
        diff,
        pr_diff,
        config,
    )
    terminal_client.run()



import logging
from dataclasses import dataclass
from timeit import default_timer
from typing import AsyncIterator, Optional

from openai.types.chat import ChatCompletionChunk

from mentat.llm_api_handler import count_tokens, model_price_per_1000_tokens
from mentat.session_context import SESSION_CONTEXT


@dataclass
class CostTracker:
    total_tokens: int = 0
    total_cost: float = 0

    last_api_call: str = ""

    def log_api_call_stats(
        self,
        num_prompt_tokens: int,
        num_sampled_tokens: int,
        model: str,
        call_time: Optional[float] = None,
        decimal_places: int = 2,
        display: bool = False,
    ) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream

        speed_and_cost_string = ""
        self.total_tokens += num_prompt_tokens + num_sampled_tokens
        if num_sampled_tokens > 0 and call_time is not None:
            tokens_per_second = num_sampled_tokens / call_time
            speed_and_cost_string += f"Speed: {tokens_per_second:.{decimal_places}f} tkns/s"
        cost = model_price_per_1000_tokens(model)
        if cost:
            prompt_cost = (num_prompt_tokens / 1000) * cost[0]
            sampled_cost = (num_sampled_tokens / 1000) * cost[1]
            call_cost = prompt_cost + sampled_cost
            self.total_cost += call_cost
            if speed_and_cost_string:
                speed_and_cost_string += " | "
            speed_and_cost_string += f"Cost: ${call_cost:.{decimal_places}f}"
        if display:
            stream.send(speed_and_cost_string, style="info")

        costs_logger = logging.getLogger("costs")
        costs_logger.info(speed_and_cost_string)
        self.last_api_call = speed_and_cost_string

    def display_last_api_call(self):
        """
        Used so that places that call the llm can print the api call stats after they finish printing everything else.
        The api call will not be logged if it gets interrupted!
        """
        ctx = SESSION_CONTEXT.get()
        if self.last_api_call:
            ctx.stream.send(self.last_api_call, style="info")

    def log_whisper_call_stats(self, seconds: float):
        self.total_cost += seconds * 0.0001

    async def response_logger_wrapper(
        self,
        prompt_tokens: int,
        response: AsyncIterator[ChatCompletionChunk],
        model: str,
    ) -> AsyncIterator[ChatCompletionChunk]:
        full_response = ""
        start_time = default_timer()
        async for chunk in response:
            # On Azure OpenAI, the first chunk streamed may contain only metadata relating to content filtering.
            if len(chunk.choices) == 0:
                continue
            full_response += chunk.choices[0].delta.content or ""
            yield chunk
        time_elapsed = default_timer() - start_time
        self.log_api_call_stats(
            prompt_tokens,
            count_tokens(full_response, model, full_message=False),
            model,
            time_elapsed,
        )



import hashlib
import logging
import os
import subprocess
from pathlib import Path
from typing import Optional, Set

from git import Repo  # type: ignore

from mentat.errors import UserError
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import is_file_text_encoded


def get_untracked_files(root: Path, paths: list[Path] = []) -> list[str]:
    """Returns untracked files. --directory flag is used to show only directories
    and not files in the output. A significant performance improvement when things
    like node_modules are present."""
    command = [
        "git",
        "ls-files",
        "--exclude-standard",
        "--others",
        "--directory",
    ] + paths
    result = subprocess.run(command, cwd=root, stdout=subprocess.PIPE)
    untracked = result.stdout.decode("utf-8").strip()
    if untracked != "":
        return untracked.split("\n")
    else:
        return []


def get_non_gitignored_files(root: Path, visited: set[Path] = set()) -> Set[Path]:
    paths = set(
        # git returns / separated paths even on windows, convert so we can remove
        # glob_excluded_files, which have windows paths on windows
        Path(os.path.normpath(p))
        for p in filter(
            lambda p: p != "",
            subprocess.check_output(
                # -c shows cached (regular) files, -o shows other (untracked/new) files
                ["git", "ls-files", "-c", "-o", "--exclude-standard"],
                cwd=root,
                text=True,
                stderr=subprocess.DEVNULL,
            ).split("\n"),
        )
        # windows-safe check if p exists in path
        if Path(root / p).exists()
    )

    file_paths: Set[Path] = set()
    # We use visited to make sure we break out of any infinite loops symlinks might cause
    visited.add(root.resolve())
    for path in paths:
        # git ls-files returns directories if the directory is itself a git project;
        # so we recursively run this function on any directories it returns.
        if (root / path).is_dir():
            if (root / path).resolve() in visited:
                continue
            file_paths.update(root / path / inner_path for inner_path in get_non_gitignored_files(root / path, visited))
        else:
            file_paths.add(path)
    return file_paths


def get_git_root_for_path(path: Path, raise_error: bool = True) -> Optional[Path]:
    if os.path.isdir(path):
        dir_path = path
    else:
        dir_path = os.path.dirname(path)
    try:
        relative_path = (
            subprocess.check_output(
                [
                    "git",
                    "rev-parse",
                    "--show-prefix",
                ],
                cwd=os.path.realpath(dir_path),
                stderr=subprocess.DEVNULL,
            )
            .decode("utf-8")
            .strip()
        )
        # --show-toplevel doesn't work in some windows environment with posix paths,
        # like msys2, so we have to use --show-prefix instead
        git_root = os.path.abspath(os.path.join(dir_path, "../" * len(Path(relative_path).parts)))
        # call realpath to resolve symlinks, so all paths match
        return Path(os.path.realpath(git_root))
    except subprocess.CalledProcessError:
        if raise_error:
            logging.error(f"File {path} isn't part of a git project.")
            raise UserError()
        else:
            return


def get_shared_git_root_for_paths(paths: list[Path]) -> Path:
    git_roots = set[Path]()
    for path in paths:
        git_root = get_git_root_for_path(path)
        if git_root is None:
            logging.error(f"File {path} isn't part of a git project.")
            raise UserError()
        git_roots.add(git_root)
    if not paths:
        git_root = get_git_root_for_path(Path(os.getcwd()))
        git_roots.add(git_root)  # pyright: ignore

    if len(git_roots) > 1:
        logging.error("All paths must be part of the same git project! Projects provided:" f" {git_roots}")
        raise UserError()
    elif len(git_roots) == 0:
        logging.error("No git projects provided.")
        raise UserError()

    return git_roots.pop()


def commit(message: str) -> None:
    """
    Commit all unstaged and staged changes
    """
    subprocess.run(["git", "add", "."])
    subprocess.run(["git", "commit", "-m", message])


def get_diff_for_file(target: str, path: Path) -> str:
    """Return commit data & diff for target versus active code"""
    # TODO: Cache git diffs and check last modified time on file
    session_context = SESSION_CONTEXT.get()

    try:
        diff_content = subprocess.check_output(
            ["git", "diff", "-U0", f"{target}", "--", path],
            cwd=session_context.cwd,
            text=True,
            stderr=subprocess.DEVNULL,
        ).strip()
        return diff_content
    except subprocess.CalledProcessError:
        logging.error(f"Error obtaining diff for commit '{target}'.")
        raise UserError()


def get_treeish_metadata(git_root: Path, target: str) -> dict[str, str]:
    try:
        commit_info = subprocess.check_output(
            ["git", "log", target, "-n", "1", "--pretty=format:%H %s"],
            cwd=git_root,
            text=True,
            stderr=subprocess.DEVNULL,
        ).strip()

        # Split the returned string into the hash and summary
        commit_hash, commit_summary = commit_info.split(" ", 1)
        return {"hexsha": commit_hash, "summary": commit_summary}
    except subprocess.CalledProcessError:
        logging.error(f"Error obtaining commit data for target '{target}'.")
        raise UserError()


def get_files_in_diff(target: str) -> list[Path]:
    """Return commit data & diff for target versus active code"""
    session_context = SESSION_CONTEXT.get()

    try:
        diff_content = subprocess.check_output(
            ["git", "diff", "--name-only", f"{target}", "--"],
            cwd=session_context.cwd,
            text=True,
            stderr=subprocess.DEVNULL,
        ).strip()
        if diff_content:
            return [Path(path) for path in diff_content.split("\n")]
        else:
            return []
    except subprocess.CalledProcessError:
        logging.error(f"Error obtaining diff for commit '{target}'.")
        raise UserError()


def check_head_exists() -> bool:
    session_context = SESSION_CONTEXT.get()

    try:
        subprocess.check_output(
            ["git", "rev-parse", "HEAD", "--"],
            cwd=session_context.cwd,
            stderr=subprocess.DEVNULL,
        )
        return True
    except subprocess.CalledProcessError:
        return False


def get_default_branch() -> str:
    session_context = SESSION_CONTEXT.get()

    try:
        # Fetch the symbolic ref of HEAD which points to the default branch
        default_branch = subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            cwd=session_context.cwd,
            text=True,
            stderr=subprocess.DEVNULL,
        ).strip()
        return default_branch
    except subprocess.CalledProcessError:
        # Handle error if needed or raise an exception
        raise Exception("Unable to determine the default branch.")


def get_git_diff(*args: str, cwd: Optional[Path] = None) -> str:
    """A wrapper on git diff that always includes new/untracked files."""
    if cwd is None:
        session_context = SESSION_CONTEXT.get()
        cwd = session_context.cwd

    repo = Repo(cwd)
    # Stage untracked files so they are included in the diff
    repo.git.add(all=True)
    diff = repo.git.diff(*args, unified=1)
    # Unstage changes again
    repo.git.reset()
    return diff + "\n" if diff else ""  # Required to form a valid .diff file


def get_hexsha_active() -> str:
    """Return the SHA-1 of the current commit."""
    session_context = SESSION_CONTEXT.get()
    cwd = session_context.cwd

    hexsha = ""
    all_files: set[Path] = get_non_gitignored_files(cwd)
    if all_files:
        hasher = hashlib.sha256()
        for file_path in sorted(all_files):
            if file_path.exists() and is_file_text_encoded(file_path):
                hasher.update(file_path.read_bytes())
        hexsha = hasher.hexdigest()
    return hexsha


# The following utilities give git information for the Mentat project itself. They are
# useful for benchmarks but shouldn't be used by the mentat module because it won't be a
# git repo when installed via pip.
def get_mentat_hexsha():
    try:
        hexsha = subprocess.check_output(
            ["git", "rev-parse", "HEAD"],
            cwd=Path(__file__).parent.parent,
            text=True,
            stderr=subprocess.DEVNULL,
        ).strip()
        return hexsha
    except subprocess.CalledProcessError:
        return ""


def get_mentat_branch():
    try:
        branch = subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            cwd=Path(__file__).parent.parent,
            text=True,
            stderr=subprocess.DEVNULL,
        ).strip()
        return branch
    except subprocess.CalledProcessError:
        return ""



from pathlib import Path

from benchmarks.exercise_runners.abstract_exercise_runner import AbstractExerciseRunner


class PythonExerciseRunner(AbstractExerciseRunner):
    def __init__(self, exercise):
        super().__init__(exercise, "py")
        self.file = Path(str(self.file).replace("-", "_"))
        self.full_path = self.dir / self.file

    def run_test(self):
        self._run_test_command(["pytest", self.dir])

    def passed(self):
        try:
            with open(self.test_output_file, "r") as f:
                lines = f.readlines()
                return "failed" not in lines[-1] and "passed" in lines[-1]
        except FileNotFoundError:
            return False



from __future__ import annotations

import base64
import io
import os
import sys
from inspect import iscoroutinefunction
from pathlib import Path
from timeit import default_timer
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    cast,
    overload,
)

import attr
import sentry_sdk
import tiktoken
from dotenv import load_dotenv
from openai import (
    APIConnectionError,
    AsyncAzureOpenAI,
    AsyncOpenAI,
    AsyncStream,
    AuthenticationError,
    AzureOpenAI,
    OpenAI,
)
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionAssistantMessageParam,
    ChatCompletionChunk,
    ChatCompletionContentPartParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
)
from openai.types.chat.completion_create_params import ResponseFormat
from PIL import Image

from mentat.errors import MentatError, ReturnToUser, UserError
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import mentat_dir_path

TOKEN_COUNT_WARNING = 32000

if TYPE_CHECKING:
    # This import is slow
    from chromadb.api.types import Embeddings


def is_test_environment():
    """Returns True if in pytest and not benchmarks"""
    benchmarks_running = os.getenv("MENTAT_BENCHMARKS_RUNNING")
    return (
        "PYTEST_CURRENT_TEST" in os.environ
        and "--benchmark" not in sys.argv
        and (not bool(benchmarks_running) or benchmarks_running == "false")
    )


def api_guard(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator that should be used on any function that calls the OpenAI API

    It does two things:
    1. Raises if the function is called in tests (that aren't benchmarks)
    2. Converts APIConnectionErrors to MentatErrors
    """

    if iscoroutinefunction(func):

        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            assert not is_test_environment(), "OpenAI call attempted in non-benchmark test environment!"
            try:
                return await func(*args, **kwargs)
            except APIConnectionError:
                raise MentatError("API connection error: please check your internet connection and" " try again.")

        return async_wrapper
    else:

        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            assert not is_test_environment(), "OpenAI call attempted in non-benchmark test environment!"
            try:
                return func(*args, **kwargs)
            except APIConnectionError:
                raise MentatError("API connection error: please check your internet connection and" " try again.")

        return sync_wrapper


# Ensures that each chunk will have at most one newline character
def chunk_to_lines(chunk: ChatCompletionChunk) -> list[str]:
    content = None
    if len(chunk.choices) > 0:
        content = chunk.choices[0].delta.content
    return ("" if content is None else content).splitlines(keepends=True)


def get_encoding_for_model(model: str) -> tiktoken.Encoding:
    try:
        # OpenAI fine-tuned models are named `ft:<base model>:<name>:<id>`. If tiktoken
        # can't match the full string, it tries to match on startswith, e.g. 'gpt-4'
        _model = model.split(":")[1] if model.startswith("ft:") else model
        return tiktoken.encoding_for_model(_model)
    except KeyError:
        return tiktoken.get_encoding("cl100k_base")


def count_tokens(message: str, model: str, full_message: bool) -> int:
    """
    Calculates the tokens in this message. Will NOT be accurate for a full prompt!
    Use prompt_tokens to get the exact amount of tokens for a prompt.
    If full_message is true, will include the extra 4 tokens used in a chat completion by this message
    if this message is part of a prompt. You do NOT want full_message to be true for a response.
    """
    encoding = get_encoding_for_model(model)
    return len(encoding.encode(message, disallowed_special=())) + (4 if full_message else 0)


def normalize_messages_for_anthropic(
    messages: list[ChatCompletionMessageParam],
) -> list[ChatCompletionMessageParam]:
    """Claude expects the chat to start with at most one system message and afterwards user and system messages to
    alternate. This method consolidates all the system messages at the beginning of the conversation into one system
    message delimited by "\n"+"-"*80+"\n and turns future system messages into user messages annotated with "System:"
    and combines adjacent assistant or user messages into one assistant or user message.
    """
    replace_non_leading_systems = list[ChatCompletionMessageParam]()
    for i, message in enumerate(messages):
        if message["role"] == "system":
            if i == 0 or messages[i - 1]["role"] == "system":
                replace_non_leading_systems.append(message)
            else:
                content = "SYSTEM: " + (message["content"] or "")
                replace_non_leading_systems.append(ChatCompletionUserMessageParam(role="user", content=content))
        else:
            replace_non_leading_systems.append(message)

    concatenate_adjacent = list[ChatCompletionMessageParam]()
    current_role: str = ""
    current_content: str = ""
    delimiter = "\n" + "-" * 80 + "\n"
    for message in replace_non_leading_systems:
        if message["role"] == current_role:
            current_content += delimiter + str(message["content"])
        else:
            if current_role == "user":
                concatenate_adjacent.append(ChatCompletionUserMessageParam(role=current_role, content=current_content))
            elif current_role == "system":
                concatenate_adjacent.append(
                    ChatCompletionSystemMessageParam(role=current_role, content=current_content)
                )
            elif current_role == "assistant":
                concatenate_adjacent.append(
                    ChatCompletionAssistantMessageParam(role=current_role, content=current_content)
                )
            current_role = message["role"]
            current_content = str(message["content"])

    if current_role == "user":
        concatenate_adjacent.append(ChatCompletionUserMessageParam(role=current_role, content=current_content))
    elif current_role == "system":
        concatenate_adjacent.append(ChatCompletionSystemMessageParam(role=current_role, content=current_content))
    elif current_role == "assistant":
        concatenate_adjacent.append(ChatCompletionAssistantMessageParam(role=current_role, content=current_content))

    return concatenate_adjacent


def prompt_tokens(messages: list[ChatCompletionMessageParam], model: str):
    """
    Returns the number of tokens used by a prompt if it was sent to OpenAI for a chat completion.
    Adapted from https://platform.openai.com/docs/guides/text-generation/managing-tokens
    """
    encoding = get_encoding_for_model(model)

    num_tokens = 0
    for message in messages:
        # every message follows <|start|>{role/name}\n{content}<|end|>\n
        # this has 5 tokens (start token, role, \n, end token, \n), but we count the role token later
        num_tokens += 4
        for key, value in message.items():
            if isinstance(value, list) and key == "content":
                value = cast(List[ChatCompletionContentPartParam], value)
                for entry in value:
                    if entry["type"] == "text":
                        num_tokens += len(encoding.encode(entry["text"]))
                    if entry["type"] == "image_url":
                        image_base64: str = entry["image_url"]["url"].split(",")[1]
                        image_bytes: bytes = base64.b64decode(image_base64)
                        image = Image.open(io.BytesIO(image_bytes))
                        size = image.size
                        # As described here: https://platform.openai.com/docs/guides/vision/calculating-costs
                        scale = min(1, 2048 / max(size))
                        size = (int(size[0] * scale), int(size[1] * scale))
                        scale = min(1, 768 / min(size))
                        size = (int(size[0] * scale), int(size[1] * scale))
                        num_tokens += 85 + 170 * ((size[0] + 511) // 512) * ((size[1] + 511) // 512)
            elif isinstance(value, str):
                num_tokens += len(encoding.encode(value))
            if key == "name":  # if there's a name, the role is omitted
                num_tokens -= 1  # role is always required and always 1 token
    num_tokens += 2  # every reply is primed with <|start|>assistant
    return num_tokens


@attr.define
class Model:
    name: str = attr.field()
    context_size: int = attr.field()
    input_cost: float = attr.field()
    output_cost: float = attr.field()
    embedding_model: bool = attr.field(default=False)


class ModelsIndex(Dict[str, Model]):
    def __init__(self, models: Dict[str, Model]):
        super().update(models)

    def _validate_key(self, key: str) -> str:
        """Try to match fine-tuned models to their base models."""
        if not super().__contains__(key) and key.startswith("ft:"):
            base_model = key.split(":")[1]  # e.g. "ft:gpt-3.5-turbo-1106:abante::8dsQMc4F"
            if super().__contains__(base_model):
                ctx = SESSION_CONTEXT.get()
                ctx.stream.send(
                    f"Using base model {base_model} for size and cost estimates.",
                    style="info",
                )
                super().__setitem__(key, attr.evolve(super().__getitem__(base_model), name=key))
                return key
        return key

    def __getitem__(self, key: str) -> Model:
        return super().__getitem__(self._validate_key(key))

    def __contains__(self, key: object) -> bool:
        return super().__contains__(self._validate_key(str(key)))


known_models = ModelsIndex(
    {
        "gpt-4-0125-preview": Model("gpt-4-0125-preview", 128000, 0.01, 0.03),
        "gpt-4-1106-preview": Model("gpt-4-1106-preview", 128000, 0.01, 0.03),
        "gpt-4-vision-preview": Model("gpt-4-vision-preview", 128000, 0.01, 0.03),
        "gpt-4": Model("gpt-4", 8192, 0.03, 0.06),
        "gpt-4-32k": Model("gpt-4-32k", 32768, 0.06, 0.12),
        "gpt-4-0613": Model("gpt-4-0613", 8192, 0.03, 0.06),
        "gpt-4-32k-0613": Model("gpt-4-32k-0613", 32768, 0.06, 0.12),
        "gpt-4-0314": Model("gpt-4-0314", 8192, 0.03, 0.06),
        "gpt-4-32k-0314": Model("gpt-4-32k-0314", 32768, 0.06, 0.12),
        "gpt-3.5-turbo-0125": Model("gpt-3.5-turbo-0125", 16385, 0.0005, 0.0015),
        "gpt-3.5-turbo-1106": Model("gpt-3.5-turbo-1106", 16385, 0.001, 0.002),
        "gpt-3.5-turbo": Model("gpt-3.5-turbo", 16385, 0.001, 0.002),
        "gpt-3.5-turbo-0613": Model("gpt-3.5-turbo-0613", 4096, 0.0015, 0.002),
        "gpt-3.5-turbo-16k-0613": Model("gpt-3.5-turbo-16k-0613", 16385, 0.003, 0.004),
        "gpt-3.5-turbo-0301": Model("gpt-3.5-turbo-0301", 4096, 0.0015, 0.002),
        "text-embedding-ada-002": Model("text-embedding-ada-002", 8191, 0.0001, 0, embedding_model=True),
    }
)


def model_context_size(model: str) -> Optional[int]:
    if model not in known_models:
        return None
    else:
        return known_models[model].context_size


def model_price_per_1000_tokens(model: str) -> Optional[tuple[float, float]]:
    """Returns (input, output) cost per 1000 tokens in USD"""
    if model not in known_models:
        return None
    else:
        return known_models[model].input_cost, known_models[model].output_cost


def get_max_tokens() -> int:
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream
    config = session_context.config

    context_size = model_context_size(config.model)
    maximum_context = config.maximum_context

    if context_size is not None and maximum_context is not None:
        return min(context_size, maximum_context)
    elif context_size is not None:
        return context_size
    elif maximum_context is not None:
        return maximum_context
    else:
        maximum_context = 4096
        # This attr has a converter from str to int
        config.maximum_context = str(maximum_context)
        stream.send(
            f"Context size for {config.model} is not known. Set maximum-context"
            " with `/config maximum_context <value>`. Using a default value of"
            f" {maximum_context}.",
            style="error",
        )
        return maximum_context


def raise_if_context_exceeds_max(tokens: int):
    ctx = SESSION_CONTEXT.get()

    max_tokens = get_max_tokens()
    if max_tokens - tokens < ctx.config.token_buffer:
        ctx.stream.send(
            f"The context size is limited to {max_tokens} tokens and your current"
            f" request uses {tokens} tokens. Please use `/exclude` to remove"
            " some files from context or `/clear` to reset the conversation",
            style="error",
        )
        raise ReturnToUser()


class LlmApiHandler:
    """Used for any functions that require calling the external LLM API"""

    def initialize_client(self):
        ctx = SESSION_CONTEXT.get()

        if not load_dotenv(mentat_dir_path / ".env"):
            load_dotenv()
        key = os.getenv("OPENAI_API_KEY")
        base_url = os.getenv("OPENAI_API_BASE")
        azure_key = os.getenv("AZURE_OPENAI_KEY")
        azure_endpoint = os.getenv("AZURE_OPENAI_ENDPOINT")

        # ChromaDB requires a sync function for embeddings
        if azure_endpoint and azure_key:
            ctx.stream.send("Using Azure OpenAI client.", style="warning")
            self.async_client = AsyncAzureOpenAI(
                api_key=azure_key,
                api_version="2023-12-01-preview",
                azure_endpoint=azure_endpoint,
            )
            self.sync_client = AzureOpenAI(
                api_key=azure_key,
                api_version="2023-12-01-preview",
                azure_endpoint=azure_endpoint,
            )
        else:
            if not key:
                if not base_url:
                    raise UserError(
                        "No OpenAI api key detected.\nEither place your key into a .env"
                        " file or export it as an environment variable."
                    )
                # If they set the base_url but not the key, they probably don't need a key, but the client requires one
                key = "fake_key"
            self.async_client = AsyncOpenAI(api_key=key, base_url=base_url)
            self.sync_client = OpenAI(api_key=key, base_url=base_url)

        try:
            self.async_client.models.list()  # Test the key
        except AuthenticationError as e:
            raise UserError(f"API gave an Authentication Error:\n{e}")

    @overload
    async def call_llm_api(
        self,
        messages: list[ChatCompletionMessageParam],
        model: str,
        stream: Literal[True],
        response_format: ResponseFormat = ResponseFormat(type="text"),
    ) -> AsyncIterator[ChatCompletionChunk]:
        ...

    @overload
    async def call_llm_api(
        self,
        messages: list[ChatCompletionMessageParam],
        model: str,
        stream: Literal[False],
        response_format: ResponseFormat = ResponseFormat(type="text"),
    ) -> ChatCompletion:
        ...

    @api_guard
    async def call_llm_api(
        self,
        messages: list[ChatCompletionMessageParam],
        model: str,
        stream: bool,
        response_format: ResponseFormat = ResponseFormat(type="text"),
    ) -> ChatCompletion | AsyncIterator[ChatCompletionChunk]:
        session_context = SESSION_CONTEXT.get()
        config = session_context.config
        cost_tracker = session_context.cost_tracker

        if "claude" in config.model:
            messages = normalize_messages_for_anthropic(messages)

        # Confirm that model has enough tokens remaining.
        tokens = prompt_tokens(messages, model)
        raise_if_context_exceeds_max(tokens)

        start_time = default_timer()
        with sentry_sdk.start_span(description="LLM Call") as span:
            span.set_tag("model", model)
            # OpenAI's API is bugged; when gpt-4-vision-preview is used, including the response format
            # at all returns a 400 error. Additionally, gpt-4-vision-preview has a max response of 30 tokens by default.
            # Until this is fixed, we have to use this workaround.
            if model == "gpt-4-vision-preview":
                response = await self.async_client.chat.completions.create(
                    model=model,
                    messages=messages,
                    temperature=config.temperature,
                    stream=stream,
                    max_tokens=4096,
                )
            else:
                # This makes it slightly easier when using the litellm proxy or models outside of OpenAI
                if response_format["type"] == "text":
                    response = await self.async_client.chat.completions.create(
                        model=model,
                        messages=messages,
                        temperature=config.temperature,
                        stream=stream,
                        max_tokens=4096,
                    )
                else:
                    response = await self.async_client.chat.completions.create(
                        model=model,
                        messages=messages,
                        temperature=config.temperature,
                        stream=stream,
                        response_format=response_format,
                        max_tokens=4096,
                    )

        # We have to cast response since pyright isn't smart enough to connect
        # the dots between stream and the overloaded create function
        if not stream:
            time_elapsed = default_timer() - start_time
            response_tokens = count_tokens(
                cast(ChatCompletion, response).choices[0].message.content or "",
                model,
                full_message=False,
            )
            cost_tracker.log_api_call_stats(tokens, response_tokens, model, time_elapsed)
        else:
            cost_tracker.last_api_call = ""
            response = cost_tracker.response_logger_wrapper(
                tokens, cast(AsyncStream[ChatCompletionChunk], response), model
            )

        return response

    @api_guard
    def call_embedding_api(self, input_texts: list[str], model: str = "text-embedding-ada-002") -> Embeddings:
        embeddings = self.sync_client.embeddings.create(input=input_texts, model=model).data
        sorted_embeddings = sorted(embeddings, key=lambda e: e.index)
        return [result.embedding for result in sorted_embeddings]

    @api_guard
    async def call_whisper_api(self, audio_path: Path) -> str:
        audio_file = open(audio_path, "rb")
        transcript = await self.async_client.audio.transcriptions.create(
            model="whisper-1",
            file=audio_file,
        )
        return transcript.text



from pathlib import Path

from mentat.utils import fetch_resource

prompts_path = "prompts"


def read_prompt(file_name: Path) -> str:
    prompt_resource = fetch_resource(prompts_path / file_name)
    with prompt_resource.open("r") as prompt_file:
        prompt = prompt_file.read()
    return prompt



from typing import List, Set

from typing_extensions import override

from mentat.command.command import Command, CommandArgument
from mentat.errors import UserError
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import get_relative_path

SEARCH_RESULT_BATCH_SIZE = 10


def _parse_include_input(user_input: str, max_num: int) -> Set[int] | None:
    nums: Set[int] = set()
    for part in user_input.split():
        left_right = part.split("-")
        if len(left_right) > 2:
            return None
        elif len(left_right) == 2:
            if not left_right[0].isdigit() or not left_right[1].isdigit():
                return None
            nums.update(range(int(left_right[0]), int(left_right[1]) + 1))
        else:
            if not part.isdigit():
                return None
            nums.add(int(part))
    nums = set(num for num in nums if num > 0 and num <= max_num)
    if not nums:
        return None
    return nums


class SearchCommand(Command, command_name="search"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_context = session_context.code_context
        config = session_context.config

        if len(args) == 0:
            stream.send("No search query specified", style="warning")
            return
        try:
            query = " ".join(args)
            results = await code_context.search(query=query)
        except UserError as e:
            stream.send(str(e), style="error")
            return

        cumulative_tokens = 0
        for i, (feature, _) in enumerate(results, start=1):
            stream.send(str(i).ljust(3), end="")
            prefix = "   "

            # TODO: The file_name was originally bolded; I decided it wasn't worth it to add a way to add
            # text attributes for this one scenario, but if we ever do add text attributes, make this bold again.
            file_name = feature.rel_path(session_context.cwd)
            stream.send(file_name, color="blue", end="")
            file_interval = feature.interval_string()
            stream.send(file_interval, color="bright_cyan", end="")

            tokens = feature.count_tokens(config.model)
            cumulative_tokens += tokens
            tokens_str = f"  ({tokens} tokens)"
            stream.send(tokens_str, color="yellow")

            if feature.name:
                name = feature.name.split(",")
                for j, n in enumerate(name):
                    stream.send(prefix, end="")
                    stream.send(f"{'└' if j == len(name) - 1 else '├'}─ ", end="")
                    stream.send(n, color="cyan")
            stream.send("")

            if i % SEARCH_RESULT_BATCH_SIZE == 0:
                # Required to avoid circular imports, but not ideal.
                from mentat.session_input import collect_user_input

                stream.send(
                    "(Y/n) for more results or to exit search mode.\nResults to"
                    ' include in context: (eg: "1 3 4" or "1-4")'
                )
                user_input: str = (await collect_user_input()).data.strip()
                while user_input.lower() not in "yn":
                    to_include = _parse_include_input(user_input, i)
                    if to_include is not None:
                        features = [results[index - 1][0] for index in to_include]
                        included_paths = code_context.include_features(features)
                        for included_path in included_paths:
                            rel_path = get_relative_path(included_path, session_context.cwd)
                            stream.send(f"{rel_path} added to context", style="success")
                    else:
                        stream.send("(Y/n)", style="input")
                    user_input: str = (await collect_user_input()).data.strip()
                if user_input.lower() == "n":
                    stream.send("Exiting search mode...", style="input")
                    break

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return [CommandArgument("required", "query")]

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        ctx = SESSION_CONTEXT.get()

        return [
            completion["display"] or completion["content"]
            for completion in ctx.auto_completer.get_file_completions(arguments[-1])
        ]

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Search files in context semantically with embeddings."



from typing import Optional

from mentat.code_feature import CodeFeature
from mentat.errors import ModelError, ReturnToUser
from mentat.feature_filters.embedding_similarity_filter import EmbeddingSimilarityFilter
from mentat.feature_filters.feature_filter import FeatureFilter
from mentat.feature_filters.llm_feature_filter import LLMFeatureFilter
from mentat.feature_filters.truncate_filter import TruncateFilter
from mentat.session_context import SESSION_CONTEXT


class DefaultFilter(FeatureFilter):
    def __init__(
        self,
        max_tokens: int,
        user_prompt: Optional[str] = None,
        expected_edits: Optional[list[str]] = None,
    ):
        self.max_tokens = max_tokens
        self.user_prompt = user_prompt or ""
        self.expected_edits = expected_edits

    async def filter(self, features: list[CodeFeature]) -> list[CodeFeature]:
        ctx = SESSION_CONTEXT.get()
        use_llm = bool(ctx.config.llm_feature_filter)

        if ctx.config.auto_context_tokens > 0 and self.user_prompt != "":
            features = await EmbeddingSimilarityFilter(self.user_prompt).filter(features)

        if use_llm:
            try:
                features = await LLMFeatureFilter(self.max_tokens, self.user_prompt, self.expected_edits).filter(
                    features
                )
            except (ModelError, ReturnToUser):
                ctx.stream.send("Feature-selection LLM response invalid. Using TruncateFilter" " instead.")
                features = await TruncateFilter(self.max_tokens, ctx.config.model).filter(features)
        else:
            features = await TruncateFilter(self.max_tokens, ctx.config.model).filter(features)

        return features
